﻿// ModuleTextProcessor.as
#pragma once
#include "..\all.h"

// Данные редактора для текстового процессора модулей. Пока хранит только флаг нужности показа списка
class ModuleEditorData : EditorData
{
    bool needQuickActivate;
    ModuleEditorData()
    {
        needQuickActivate = true;
    }
};

// Контексты срабатывания подсказки снегопата
enum ActivateModes {
    modeNone,       // нет контекста
    modeQuote,      // внутри кавычек
    modeName,       // обычный режим, в идентификаторе
    modePreproc     // в инструкции препроцессора
};

string getTextLine(TextManager&& tm, uint line)
{
    v8string wline;
    tm.getLine(line, wline);
    return wline.str;
}

bool my_is_alpha(wchar_t symbol)
{
    return mem::byte[symbolClassTable + symbol] == symbAlpha;
}

bool my_is_alnum(wchar_t symbol)
{
    return mem::byte[symbolClassTable + symbol] == symbAlpha || (symbol >= '0' && symbol <= '9');
}

bool is_first_name_symbol(wchar_t symbol)
{
    return my_is_alpha(symbol) || '_' == symbol;
}

bool is_name_symbol(wchar_t symbol)
{
    return my_is_alnum(symbol) || '_' == symbol;
}

// Выяснить контекст вызова подсказки
// также возвращает начало текущей строки до каретки и последнюю лексему в ней.
ActivateModes getActivateMode(TextManager&& tm, const TextPosition&in caretPos, string& beginOfLine, string& lastLexem)
{
    uint symbolsToCaret = caretPos.col - 1;
    beginOfLine = getTextLine(tm, caretPos.line).rtrim("\r\n").padRight(' ', symbolsToCaret);
    beginOfLine.setLength(symbolsToCaret);

    lex_provider source(beginOfLine.cstr);
    lexem lex;
    do {
        source.next(lex);
    } while (!(lex.isEmpty() || source.atEnd()));
    
    if (lexPreproc == lex.type)
        return modePreproc;
    else if (lexRemark == lex.type || lexNumber == lex.type || lexDateOpen == lex.type || lexDate == lex.type)
        return modeNone;
    else if (lexQuoteOpen == lex.type)// Внутри строки
        return modeQuote;
    else {
        if (lexName == lex.type || lexDirective == lex.type)
            lastLexem = lex.text;
        return modeName;
    }
}

class ModuleTextProcessor : TextProcessor
{
    void connectEditor(TextWnd&& editor) { &&editor.editorData = ModuleEditorData(); }
    void afterChar(TextWnd&& editor, wchar_t symbol)
    {
        if (editor !is activeTextWnd)
            return;
        bool bItsNameSymbol = is_name_symbol(symbol);
        ModuleEditorData&& pData = cast<ModuleEditorData>(editor.editorData);
        if (!bItsNameSymbol)
            pData.needQuickActivate = true;
        TextPosition caretPos;
        editor.ted.getCaretPosition(caretPos, false);
        if ('\r' == symbol) {
            smartEnter(caretPos, editor);
            return;
        }
        string lineBegin, lastLexem;
        ActivateModes mode = getActivateMode(editor.textDoc.tm, caretPos, lineBegin, lastLexem);
        if (modeName == mode && doAutoReplace(symbol, lineBegin, caretPos, editor))
            return;
        /*
        IProfile* pProf = get_process_service<IProfile>();
        int qaChars;
        if (!pProf.getValue(SETTINGS_QACHARS, qaChars))
            qaChars = 1;
        */
        uint qaChars = 2;

        if (modeName == mode && ((bItsNameSymbol && pData.needQuickActivate && lastLexem.length >= qaChars))) {
            activateInModule(editor, lastLexem, caretPos, mode, symbol);
        }/* else if (modeQuote == mode)
            activateInString(pEditor, caretPos);
        else if (modePreproc == mode)
            activateInPreproc(pEditor, caretPos);
        else if ('(' == symbol || ',' == symbol)
            ParamShow::get().activate();
        */
    }
    bool onKeyDown(TextWnd&& editor, uint wParam, uint lParam)
    {
        switch (wParam) {
        case VK_LEFT:       case VK_RIGHT:      case VK_UP:
        case VK_DOWN:       case VK_PRIOR:      case VK_NEXT:
        case VK_HOME:       case VK_END:        case VK_RETURN:
        case VK_BACK:
            cast<ModuleEditorData>(editor.editorData).needQuickActivate = true;
            break;
        case VK_DELETE:
            //m_pModElements.setNeedReparse();
            break;
        case VK_ESCAPE:
            //ParamShow::get().hide();
            break;
        }
        if ((GetKeyState(VK_MENU) & 0x8000) != 0) { // Нажата Alt
            string replSymb = "";
            switch (wParam) {
            case VK_OEM_4:  // [{
                if ((GetKeyState(VK_SHIFT) & 0x8000) != 0)
                    replSymb = "{¦}";
                else
                    replSymb = "[¦]";
                break;
            case VK_OEM_6:  // ]}
                if ((GetKeyState(VK_SHIFT) & 0x8000) != 0)
                    replSymb = "}";
                else
                    replSymb = "]";
                break;
            case VK_OEM_COMMA:
                replSymb = "<";
                break;
            case VK_OEM_PERIOD:
                replSymb = ">";
                break;
            case '3':
                replSymb = "#";
                break;
            case '7':
                replSymb = "&";
                break;
            case '4':
                replSymb = "$";
                break;
            case '5':
                {
                    IntelliSite&& site = getIntelliSite();
                    //addTypeNames(site);
                    //site.addItemGroup(v8stock[stockGlobalProc][langRus]);
                    //site.addItemGroup(v8stock[stockSysEnums][langRus]);
                    //site.addItemGroup(v8stock[stockGlobalProps][langRus]);
                    addAllStock(site);
                    site.show(editor, "");
                }
                return true;
            }
            if (!replSymb.isEmpty()) {
                insertInSelection(editor.ted, editor.textDoc.tm, editor.textDoc.itm, replSymb, true, false);
                return true;
            }
        }
        return false;
    }
    void itemInserted(TextWnd&& editor, SmartBoxInsertableItem&& item)
    {
        cast<ModuleEditorData>(editor.editorData).needQuickActivate = true;
    }

    // Вызывается при нажатии Enter'а.
    void smartEnter(TextPosition& caretPos, TextWnd&& editor)
    {
        v8string lineOver;
        v8string line;
        editor.textDoc.tm.getLine(caretPos.line - 1, lineOver);
        lex_provider lexSrc(lineOver.cstr);
        lexem lex;
        while (lexSrc.next(lex)) {
            if (lexQuoteOpen == lex.type) {
                editor.textDoc.tm.getLine(caretPos.line, line);
                if (line.str.ltrim().substr(0, 1) != "|") {
                    // Надо вставить |
                    // Получим пробельные символы для автоотступа
                    string str = lineOver.str.ltrim();
                    string indent;
                    if (str[0] == '|' || str[0] == '\"')
                        indent = str.substr(1).rtrim("\r\n").match(indentRex).text(0, 0);
                    insertInSelection(editor.ted, editor.textDoc.tm, editor.textDoc.itm, "|" + indent, true, false);
                } else {
                    caretPos.col++;
                    editor.ted.setCaretPosition(caretPos, false);
                }
                break;
            } else if (lexRemark == lex.type && (GetKeyState(VK_SHIFT) & 0x8000) != 0) {
                insertInSelection(editor.ted, editor.textDoc.tm, editor.textDoc.itm, "// ", true, false);
                break;
            }
        }
    }

    bool doAutoReplace(wchar_t symbol, const string&in lineBegin, TextPosition&in caretPos, TextWnd&& pTxtWnd)
    {
        //if (!bAutoReplaces)
        //    return FALSE;
        //	Делаем автозамены: ? ++ -- += -= *= /= %= 
        uint len = lineBegin.length;
        /*
        if ('?' == symbol) {
            IntelliSite& is = IntelliSite::get();
            is.addItem(Delimeters::getDelimeter(Delimeters::question));
            is.show(CString(), pTxtWnd);
            return TRUE;
        }*/
        // Для начала определим, надо ли вообще делать автозамену
        if (len < 3)
            return false;
        wchar_t prevSymbol = lineBegin[len - 2];
        string replSymbol1 = "", replSymbol2 = "";
        if ('+' == symbol && '+' == prevSymbol) {
            replSymbol1 = "+";
            replSymbol2 = "1";
        } else if ('-' == symbol && '-' == prevSymbol) {
            replSymbol1 = "-";
            replSymbol2 = "1";
        } else if ('=' == symbol && "+-*/%".find(prevSymbol) >= 0) {
            replSymbol1.insert(0, prevSymbol);
            replSymbol2.insert(0, symbCaret);
        } else
            return false;
        // Дальше надо определить, нужно ли добавлять пробелы и получить идентификатор для вставки
        string spaceSymbol = "";
        int endOfName = len - 3;
        // ищем конец имени перед заменой
        while (endOfName >= 0 && is_space(lineBegin[endOfName])) {
            endOfName--;
            spaceSymbol = " ";
        }
        if (endOfName < 0)
            return false;
        // ищем начало имени
        int startOfName = endOfName;
        while (startOfName >= 0 && !is_space(lineBegin[startOfName]) && lineBegin[startOfName] != ';')
            startOfName--;
        startOfName++;
        if (startOfName > endOfName || !is_first_name_symbol(lineBegin[startOfName]))
            return false;
        // выделяем  имя
        string id = lineBegin.substr(startOfName, endOfName - startOfName + 1);
        // формируем строку замены
        string replace = "=" + spaceSymbol + id + spaceSymbol + replSymbol1 + spaceSymbol + replSymbol2;
        // получаем часть строки после положения каретки, и проверяем, не пустая ли она
        // регэксп в конце нужен, так как в юникоде помимо пробела и табуляции могут быть и другие
        // whitespace символы. В оригинале снегопата тут в цикле каждый символ проверяется на whitespace
        string endOfLine = getTextLine(pTxtWnd.textDoc.tm, caretPos.line).substr(caretPos.col - 1).rtrim(" \t\r\n").replace(indentRex, "");
        if (endOfLine.isEmpty())    // Если остаток строки пустой,
            replace += ";";         // добавим запяточку
        // Собственно, сделаем вставку
        TextPosition tpStart = caretPos;
        tpStart.col -= 2;
        pTxtWnd.ted.setSelection(tpStart, caretPos, false, false);
        insertInSelection(pTxtWnd.ted, pTxtWnd.textDoc.tm, pTxtWnd.textDoc.itm, replace, true, false);
        return true;
    }
    void activateInModule(TextWnd&& editor, const string&in lastLexem, TextPosition& caretPos, ActivateModes mode, wchar_t symbol)
    {
        cast<ModuleEditorData>(editor.editorData).needQuickActivate = false;
        // Пока только так
        // cast<ModuleEditorData>(editor.editorData).needQuickActivate = false;
        showV8Assist();
    }
};

void showProfileFolder(IProfileFolder&& folder, const string& level)
{
    string indent = level;
    Message(indent + "[" + folder.getName() + "]");
    indent += "\t";
    Value val;
    v8string vname, vval;
    for (uint k = 0, m = folder.valuesCount(); k < m ; k++) {
        folder.getValueKeyAt(k, vname);
        folder.getValueAt(k, val);
        val.pValue.getString(vval);
        Message(indent + vname + " = " + vval);
    }
    for (uint k = 0, m = folder.foldersCount(); k < m; k++) {
        IProfileFolder&& subFolder;
        folder.getFolderAt(subFolder, k);
        showProfileFolder(subFolder, indent);
    }
}

void showV8Assist()
{
    //showProfileFolder(getProfileRoot(), "");
    CommandID assist(cmdFrameGroup, cmdFrameShowAssist);
    if (commandState(assist) & cmdStateEnabled != 0)
        sendCommandToMainFrame(assist);
}
