﻿/*  lexer.as
   Лексеры и парсеры
   Как скриптовые, так и встроенные в снегопат
*/
#pragma once
#include "..\all.h"

// Типы лексем, используемые встроенным в снегопат парсером
enum LexemTypes
{
    endOfText = 0,
    lexUnknown = 0,
    lexRemark,      // комментарий
    lexQuote,       // в кавычках
    lexQuoteOpen,   // в открытых кавычках (нет завершающей кавычки)
    lexDate,        // дата
    lexDateOpen,    // дата без завершающего апострофа
    lexNumber,      // число
    lexPreproc,     // инструкция препроцессора
    lexDirective,   // директива выполнения
    lexLabel,       // метка
    lexLPar,        // (
    lexRPar,        // )
    lexLBrt,        // [
    lexRBrt,        // ]
    lexEqual,       // =
    lexComma,       // ,
    lexSemicolon,   // ;
    lexPlus,        // +
    lexMinus,       // -
    lexMult,        // *
    lexDivide,      // /
    lexMod,         // %
    lexQuestion,    // ?
    lexPeriod,      // .
    lexLess,        // <
    lexLessEq,      // <=
    lexGrat,        // >
    lexGratEq,      // >=
    lexNotEq,       // <>
    lexName,        // идентификатор
    // Ключевые слова
    kwIf,
    kwThen,
    kwElsIf,
    kwEndIf,
    kwElse,
    kwFor,
    kwEach,
    kwIn,
    kwTo,
    kwWhile,
    kwDo,
    kwEndDo,
    kwProcedure,
    kwFunction,
    kwEndProcedure,
    kwEndFunction,
    kwVar,
    kwGoto,
    kwReturn,
    kwContinue,
    kwBreak,
    kwAnd,
    kwOr,
    kwNot,
    kwTry,
    kwExcept,
    kwRaise,
    kwEndTry,
    kwNew,
    kwExecute,
    kwTrue,
    kwFalse,
    kwAddHandler,
    kwRemoveHandler,
    kwExport,
    kwNull,
    kwUndefined,
    kwVal,
    // Спец-токены для синтакс-парсера, добавляются в начало текста для задания под-алгоритма парсера
    // допустимы только методы. Указывается, если анализ идет в общем модуле,
    // или начинается с середины модуля между методами
    onlyMeths,
    // Указывается, если анализ начинается с части модуля, за которой нет методов
    onlyMethsAndStatements,
    // Указывается при анализе с начала модуля, когда после уже есть методы
    onlyVarsAndMethods,
    kwCount = onlyMeths - lexName - 1,
};

// Кое-какие регэкспы. Пока нормальный парсер не портировал, используем их
RegExp quotesRex("""["\|][^"]*(?:["\n]|$)""");    // Для определения, заканчивается ли строка текста открытым литералом (т.е. без завершающей кавычки)
RegExp indentRex("^\\s*");  // Получение отступа строки

// Проверка, завершается ли строка текста открытым литералом
bool isLineEndWithOpenQuote(const string& line)
{
    auto res = line.match(quotesRex);
    return res.matches > 0 && (res.len(res.matches - 1, 0) == 1 || res.text(res.matches - 1, 0).substr(-1) != "\"");
}

// Функция для чтения содержимого текстового файла.
// Чтение осуществляется с помощью встроенного объекта 1С "ТекстовыйДокумент".
// Можно указать кодировку текста.
bool readTextFile(v8string& result, const string& path, const string& encoding = "")
{
    IContext&& textDoc;
    currentProcess().createByClsid(CLSID_TxtEdtDoc, IID_IContext, textDoc);
    if (textDoc is null)
        return false;
    IContextDef&& pCtxDef = textDoc;
    int methPos = pCtxDef.findMethod("Read");
    if (methPos < 0)
        return false;
    ValueParamsVector params(3);
    params.values[0] = path;
    
    if (encoding.isEmpty())
        pCtxDef.getParamDefValue(methPos, 1, params.values[1]);
    else
        params.values[1] = encoding;
    pCtxDef.getParamDefValue(methPos, 2, params.values[2]);
   
    Value retVal;
    textDoc.callMeth(methPos, params.retVal, params.args);
    
    ITextManager&& itm = cast<IUnknown>(textDoc);
    itm.getTextManager().save(result);
    return true;
}

class ValueParamsVector {
    Vector args;
    Value[]&& values;
    Value retVal;

    ValueParamsVector(uint argsCount)
    {
        &&values = Value[](argsCount);
        uint mem = 4 * argsCount;
        args.start = malloc(mem);
        args.allocked = args.end = args.start + mem;
        for (uint i = 0; i < argsCount; i++)
            mem::dword[args.start + i * 4] = values[i].self;
    }
};